@@ -1,6 +1,6 @@
-{-# LANGUAGE CPP #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE TypeApplications #-}
 -- | This module provides support for CPP, interpreter directives and line
 -- pragmas.
 module Language.Haskell.GHC.ExactPrint.Preprocess
@@ -12,58 +12,36 @@ module Language.Haskell.GHC.ExactPrint.Preprocess
 
    , CppOptions(..)
    , defaultCppOptions
+   , showErrorMessages
    ) where
 
 import qualified GHC            as GHC hiding (parseModule)
 
-#if __GLASGOW_HASKELL__ >= 900
 import qualified Control.Monad.IO.Class as GHC
 import qualified GHC.Data.Bag          as GHC
 import qualified GHC.Data.FastString   as GHC
 import qualified GHC.Data.StringBuffer as GHC
+-- import qualified GHC.Driver.Config     as GHC
+import qualified GHC.Driver.Config.Parser     as GHC
+import qualified GHC.Driver.Env        as GHC
+import qualified GHC.Driver.Errors.Types        as GHC
 import qualified GHC.Driver.Phases     as GHC
 import qualified GHC.Driver.Pipeline   as GHC
--- import qualified GHC.Driver.Session    as GHC
-import qualified GHC.Driver.Types      as GHC
 import qualified GHC.Fingerprint.Type  as GHC
-import qualified GHC.Utils.Fingerprint as GHC
+-- import qualified GHC.Parser.Errors.Ppr as GHC
 import qualified GHC.Parser.Lexer      as GHC
 import qualified GHC.Settings          as GHC
+import qualified GHC.Types.SourceError as GHC
 import qualified GHC.Types.SrcLoc      as GHC
-import qualified GHC.Utils.Error       as GHC
+import qualified GHC.Types.SourceFile  as GHC
+import qualified GHC.Types.Error  as GHC
+-- import qualified GHC.Utils.Error       as GHC
+import qualified GHC.Utils.Fingerprint as GHC
 import GHC.Types.SrcLoc (mkSrcSpan, mkSrcLoc)
 import GHC.Data.FastString (mkFastString)
-#else
-import qualified Bag            as GHC
-import qualified DriverPhases   as GHC
-import qualified DriverPipeline as GHC
-import qualified DynFlags       as GHC
-import qualified ErrUtils       as GHC
-import qualified FastString     as GHC
-import qualified HscTypes       as GHC
-import qualified Lexer          as GHC
-import qualified MonadUtils     as GHC
-import qualified SrcLoc         as GHC
-import qualified StringBuffer   as GHC
-import SrcLoc (mkSrcSpan, mkSrcLoc)
-import FastString (mkFastString)
-#endif
 
-#if (__GLASGOW_HASKELL__ > 808) && (__GLASGOW_HASKELL__ < 900)
-import qualified Fingerprint    as GHC
-import qualified ToolSettings   as GHC
-#endif
-
-
-#if __GLASGOW_HASKELL__ > 808
-#else
-import Control.Exception
-#endif
-import Data.List hiding (find)
+import Data.List (isPrefixOf, intercalate)
 import Data.Maybe
-#if __GLASGOW_HASKELL__ <= 800
-import Language.Haskell.GHC.ExactPrint.GhcInterim (commentToAnnotation)
-#endif
 import Language.Haskell.GHC.ExactPrint.Types
 import Language.Haskell.GHC.ExactPrint.Utils
 import qualified Data.Set as Set
@@ -88,29 +66,29 @@ defaultCppOptions = CppOptions [] [] []
 
 -- ---------------------------------------------------------------------
 -- | Remove GHC style line pragams (@{-# LINE .. #-}@) and convert them into comments.
-stripLinePragmas :: String -> (String, [Comment])
+stripLinePragmas :: String -> (String, [GHC.LEpaComment])
 stripLinePragmas = unlines' . unzip . findLines . lines
   where
     unlines' (a, b) = (unlines a, catMaybes b)
 
-findLines :: [String] -> [(String, Maybe Comment)]
+findLines :: [String] -> [(String, Maybe GHC.LEpaComment)]
 findLines = zipWith checkLine [1..]
 
-checkLine :: Int -> String -> (String, Maybe Comment)
+checkLine :: Int -> String -> (String, Maybe GHC.LEpaComment)
 checkLine line s
   |  "{-# LINE" `isPrefixOf` s =
        let (pragma, res) = getPragma s
            size   = length pragma
            mSrcLoc = mkSrcLoc (mkFastString "LINE")
            ss     = mkSrcSpan (mSrcLoc line 1) (mSrcLoc line (size+1))
-       in (res, Just $ mkComment pragma (rs ss))
+       in (res, Just $ mkLEpaComment pragma (GHC.spanAsAnchor ss) (GHC.realSrcSpan ss))
   -- Deal with shebang/cpp directives too
   -- x |  "#" `isPrefixOf` s = ("",Just $ Comment ((line, 1), (line, length s)) s)
   |  "#!" `isPrefixOf` s =
     let mSrcLoc = mkSrcLoc (mkFastString "SHEBANG")
         ss = mkSrcSpan (mSrcLoc line 1) (mSrcLoc line (length s))
     in
-    ("",Just $ mkComment s (rs ss))
+    ("",Just $ mkLEpaComment s (GHC.spanAsAnchor ss) (GHC.realSrcSpan ss))
   | otherwise = (s, Nothing)
 
 getPragma :: String -> (String, String)
@@ -129,46 +107,38 @@ getPragma s@(x:xs)
 getCppTokensAsComments :: GHC.GhcMonad m
                        => CppOptions  -- ^ Preprocessor Options
                        -> FilePath    -- ^ Path to source file
-                       -> m [Comment]
+                       -> m [GHC.LEpaComment]
 getCppTokensAsComments cppOptions sourceFile = do
   source <- GHC.liftIO $ GHC.hGetStringBuffer sourceFile
   let startLoc = GHC.mkRealSrcLoc (GHC.mkFastString sourceFile) 1 1
-  (_txt,strSrcBuf,flags2) <- getPreprocessedSrcDirectPrim cppOptions sourceFile
-  -- #ifdef tokens
+  (_txt,strSrcBuf,flags2') <- getPreprocessedSrcDirectPrim cppOptions sourceFile
+  let flags2 = GHC.initParserOpts flags2'
+  -- hash-ifdef tokens
   directiveToks <- GHC.liftIO $ getPreprocessorAsComments sourceFile
-  -- Tokens without #ifdef
+  -- Tokens without hash-ifdef
   nonDirectiveToks <- tokeniseOriginalSrc startLoc flags2 source
-  case GHC.lexTokenStream strSrcBuf startLoc flags2 of
+  case GHC.lexTokenStream flags2 strSrcBuf startLoc of
         GHC.POk _ ts ->
                do
                   let toks = GHC.addSourceToTokens startLoc source ts
                       cppCommentToks = getCppTokens directiveToks nonDirectiveToks toks
                   return $ filter goodComment
-#if __GLASGOW_HASKELL__ >= 900
-                         $  map (tokComment . GHC.commentToAnnotation . toRealLocated . fst) cppCommentToks
-#elif __GLASGOW_HASKELL__ > 800
-                         $  map (tokComment . GHC.commentToAnnotation . fst) cppCommentToks
-#else
-                         $  map (tokComment . commentToAnnotation . fst) cppCommentToks
-#endif
-#if __GLASGOW_HASKELL__ > 808
-        GHC.PFailed pst -> parseError flags2 pst
-#elif __GLASGOW_HASKELL__ >= 804
-        GHC.PFailed _ sspan err -> parseError flags2 sspan err
-#else
-        GHC.PFailed sspan err -> parseError flags2 sspan err
-#endif
+                         $  map (GHC.commentToAnnotation . toRealLocated . fst) cppCommentToks
+        GHC.PFailed pst -> parseError pst
 
-goodComment :: Comment -> Bool
-goodComment (Comment "" _ _) = False
-goodComment _              = True
+
+goodComment :: GHC.LEpaComment -> Bool
+goodComment c = isGoodComment (tokComment c)
+  where
+    isGoodComment :: [Comment] -> Bool
+    isGoodComment []                 = False
+    isGoodComment [Comment "" _ _ _] = False
+    isGoodComment _                  = True
 
 
-#if __GLASGOW_HASKELL__ >= 900
 toRealLocated :: GHC.Located a -> GHC.RealLocated a
 toRealLocated (GHC.L (GHC.RealSrcSpan s _) x) = GHC.L s              x
 toRealLocated (GHC.L _ x)                     = GHC.L badRealSrcSpan x
-#endif
 
 -- ---------------------------------------------------------------------
 
@@ -205,7 +175,7 @@ getCppTokens directiveToks origSrcToks postCppToks = toks
     missingAsComments = map mkCommentTok missingToks
       where
         mkCommentTok :: (GHC.Located GHC.Token,String) -> (GHC.Located GHC.Token,String)
-        mkCommentTok (GHC.L l _,s) = (GHC.L l (GHC.ITlineComment s),s)
+        mkCommentTok (GHC.L l _,s) = (GHC.L l (GHC.ITlineComment s (makeBufSpan l)),s)
 
     toks = mergeBy locFn directiveToks missingAsComments
 
@@ -213,19 +183,13 @@ getCppTokens directiveToks origSrcToks postCppToks = toks
 
 tokeniseOriginalSrc ::
   GHC.GhcMonad m
-  => GHC.RealSrcLoc -> GHC.DynFlags -> GHC.StringBuffer
+  => GHC.RealSrcLoc -> GHC.ParserOpts -> GHC.StringBuffer
   -> m [(GHC.Located GHC.Token, String)]
 tokeniseOriginalSrc startLoc flags buf = do
   let src = stripPreprocessorDirectives buf
-  case GHC.lexTokenStream src startLoc flags of
+  case GHC.lexTokenStream flags src startLoc of
     GHC.POk _ ts -> return $ GHC.addSourceToTokens startLoc src ts
-#if __GLASGOW_HASKELL__ > 808
-    GHC.PFailed pst -> parseError flags pst
-#elif __GLASGOW_HASKELL__ >= 804
-    GHC.PFailed _ sspan err -> parseError flags sspan err
-#else
-    GHC.PFailed sspan err -> parseError flags sspan err
-#endif
+    GHC.PFailed pst -> parseError pst
 
 -- ---------------------------------------------------------------------
 
@@ -259,27 +223,27 @@ getPreprocessedSrcDirectPrim cppOptions src_fn = do
   hsc_env <- GHC.getSession
   let dfs = GHC.hsc_dflags hsc_env
       new_env = hsc_env { GHC.hsc_dflags = injectCppOptions cppOptions dfs }
-#if __GLASGOW_HASKELL__ >= 808
-  -- (dflags', hspp_fn) <-
   r <- GHC.liftIO $ GHC.preprocess new_env src_fn Nothing (Just (GHC.Cpp GHC.HsSrcFile))
   case r of
-    Left err -> error $ showErrorMessages err
+    Left err -> error $ showErrorMessages $ fmap GHC.GhcDriverMessage err
     Right (dflags', hspp_fn) -> do
       buf <- GHC.liftIO $ GHC.hGetStringBuffer hspp_fn
       txt <- GHC.liftIO $ readFileGhc hspp_fn
       return (txt, buf, dflags')
-#else
-  (dflags', hspp_fn) <-
-      GHC.liftIO $ GHC.preprocess new_env (src_fn, Just (GHC.Cpp GHC.HsSrcFile))
-  buf <- GHC.liftIO $ GHC.hGetStringBuffer hspp_fn
-  txt <- GHC.liftIO $ readFileGhc hspp_fn
-  return (txt, buf, dflags')
-#endif
 
-#if __GLASGOW_HASKELL__ >= 808
 showErrorMessages :: GHC.ErrorMessages -> String
-showErrorMessages msgs = intercalate "\n" $ map show $ GHC.bagToList msgs
-#endif
+showErrorMessages msgs = intercalate "\n"
+    $ map (show @(GHC.MsgEnvelope GHC.DiagnosticMessage) . fmap toDiagnosticMessage)
+    $ GHC.bagToList
+    $ GHC.getErrorMessages msgs
+
+-- | Show Error Messages relies on show instance for MsgEnvelope DiagnosticMessage
+-- We convert a known Diagnostic into this generic version
+toDiagnosticMessage :: GHC.Diagnostic e => e -> GHC.DiagnosticMessage
+toDiagnosticMessage msg = GHC.DiagnosticMessage { diagMessage = GHC.diagnosticMessage msg
+                                                , diagReason  = GHC.diagnosticReason  msg
+                                                , diagHints   = GHC.diagnosticHints   msg
+                                                }
 
 injectCppOptions :: CppOptions -> GHC.DynFlags -> GHC.DynFlags
 injectCppOptions CppOptions{..} dflags =
@@ -290,7 +254,6 @@ injectCppOptions CppOptions{..} dflags =
     mkInclude = ("-include" ++)
 
 
-#if __GLASGOW_HASKELL__ > 808
 addOptP :: String -> GHC.DynFlags -> GHC.DynFlags
 addOptP   f = alterToolSettings $ \s -> s
           { GHC.toolSettings_opt_P   = f : GHC.toolSettings_opt_P s
@@ -302,13 +265,6 @@ alterToolSettings f dynFlags = dynFlags { GHC.toolSettings = f (GHC.toolSettings
 fingerprintStrings :: [String] -> GHC.Fingerprint
 fingerprintStrings ss = GHC.fingerprintFingerprints $ map GHC.fingerprintString ss
 
-#else
-addOptP :: String -> GHC.DynFlags -> GHC.DynFlags
-addOptP   f = alterSettings (\s -> s { GHC.sOpt_P   = f : GHC.sOpt_P s})
-
-alterSettings :: (GHC.Settings -> GHC.Settings) -> GHC.DynFlags -> GHC.DynFlags
-alterSettings f dflags = dflags { GHC.settings = f (GHC.settings dflags) }
-#endif
 -- ---------------------------------------------------------------------
 
 -- | Get the preprocessor directives as comment tokens from the
@@ -319,7 +275,7 @@ getPreprocessorAsComments srcFile = do
   let directives = filter (\(_lineNum,line) -> line /= [] && head line == '#')
                     $ zip [1..] (lines fcontents)
 
-  let mkTok (lineNum,line) = (GHC.L l (GHC.ITlineComment line),line)
+  let mkTok (lineNum,line) = (GHC.L l (GHC.ITlineComment line (makeBufSpan l)),line)
        where
          start = GHC.mkSrcLoc (GHC.mkFastString srcFile) lineNum 1
          end   = GHC.mkSrcLoc (GHC.mkFastString srcFile) lineNum (length line)
@@ -328,20 +284,21 @@ getPreprocessorAsComments srcFile = do
   let toks = map mkTok directives
   return toks
 
+makeBufSpan :: GHC.SrcSpan -> GHC.PsSpan
+makeBufSpan ss = pspan
+  where
+    bl = GHC.BufPos 0
+    pspan = GHC.PsSpan (GHC.realSrcSpan ss) (GHC.BufSpan bl bl)
+
 -- ---------------------------------------------------------------------
 
-#if __GLASGOW_HASKELL__ > 808
-parseError :: (GHC.MonadIO m) => GHC.DynFlags -> GHC.PState -> m b
-parseError dflags pst = do
+parseError :: (GHC.MonadIO m) => GHC.PState -> m b
+parseError pst = do
      let
        -- (warns,errs) = GHC.getMessages pst dflags
      -- throw $ GHC.mkSrcErr (GHC.unitBag $ GHC.mkPlainErrMsg dflags sspan err)
-     GHC.throwErrors (GHC.getErrorMessages pst dflags)
-#else
-parseError :: GHC.DynFlags -> GHC.SrcSpan -> GHC.MsgDoc -> m b
-parseError dflags sspan err = do
-     throw $ GHC.mkSrcErr (GHC.unitBag $ GHC.mkPlainErrMsg dflags sspan err)
-#endif
+     -- GHC.throwErrors (fmap GHC.mkParserErr (GHC.getErrorMessages pst))
+     GHC.throwErrors (fmap GHC.GhcPsMessage (GHC.getPsErrorMessages pst))
 
 -- ---------------------------------------------------------------------
 
@@ -373,4 +330,3 @@ mergeBy cmp (allx@(x:xs)) (ally@(y:ys))
         -- Someone please put this code out of its misery.
     | (x `cmp` y) <= EQ = x : mergeBy cmp xs ally
     | otherwise = y : mergeBy cmp allx ys
-
